"""Handle data processing and laoding."""

from typing import Any, Optional, Tuple
from pathlib import Path
import json
import logging
import gzip
import pickle as pkl

import numpy as np
from sklearn.utils import compute_class_weight
from datasets import Dataset, DatasetDict, load_from_disk, disable_progress_bars


_SEED = 42
_DATA_DIR = (Path(__file__).parent / "datasets").resolve()
_DATASETS_NAMES = [d.name for d in _DATA_DIR.iterdir() if d.is_dir()]


def load_data(
    dataset_name: str,
    predefined_split_info: Optional[dict[str, Any]] = None,
    use_validation: bool = True,
    test_size: float = 0.15,
) -> Tuple[Dataset, Dataset, Dataset]:
    disable_progress_bars()
    assert (
        dataset_name in _DATASETS_NAMES
    ), f"Dataset {dataset_name} not found. Must be one of {_DATASETS_NAMES}"
    dataset_path = _DATA_DIR / dataset_name / "hf_dataset"
    try:
        if predefined_split_info:
            logging.info(f"Loading dataset {dataset_name} from predefined splits")
            train_data, test_data, val_data = _load_data_from_predefined_splits(
                dataset_name, predefined_split_info, use_validation
            )
        else:
            logging.info(f"Loading dataset {dataset_name} from HuggingFace dataset")
            train_data, test_data, val_data = _load_data_from_hf_splits(
                dataset_name, use_validation
            )
    except FileNotFoundError:
        logging.warning(
            f"Dataset {dataset_name} not found in {dataset_path}. Processing raw data."
        )
        train_data, test_data, val_data = preprocess_data(dataset_name, test_size, use_validation)
        logging.info(f"Saving processed dataset to {dataset_path}")
        if val_data is not None:
            dataset_dict = DatasetDict(
                {"train": train_data, "test": test_data, "validation": val_data}
            )
        else:
            dataset_dict = DatasetDict({"train": train_data, "test": test_data})
        dataset_dict.save_to_disk(dataset_path)

    return train_data, test_data, val_data


def _load_data_from_hf_splits(dataset_name, use_validation):
    dataset_path = _DATA_DIR / dataset_name / "hf_dataset"
    dataset_dict = load_from_disk(dataset_path)
    train_data = dataset_dict["train"]
    test_data = dataset_dict["test"]
    if use_validation and "validation" not in dataset_dict:
        raise FileNotFoundError
    val_data = dataset_dict["validation"] if use_validation else None
    return train_data, test_data, val_data


def _load_data_from_predefined_splits(dataset_name, predefined_split_info, use_validation):
    """Load dataset from pre-split txt files containing proteins IDs."""
    dataset_dir = _DATA_DIR / dataset_name
    assert dataset_dir.exists(), f"Dataset {dataset_name} not found in {dataset_dir}"
    sequences = _unpickle(dataset_dir / "sequences.pkl")
    labels = _unpickle(dataset_dir / "labels.pkl")

    data = {"id": [], "sequence": [], "labels": []}
    for key in labels.keys():
        data["id"].append(key)
        data["sequence"].append("".join(sequences[key]))
        data["labels"].append(labels[key])
        if len(sequences[key]) != len(labels[key]):
            logging.warning(f"Sequence and label lengths do not match for {key}")
    data = Dataset.from_dict(data)
    splits_file = dataset_dir / predefined_split_info["splits_file"]  # relative path to JSON file
    if not splits_file.exists():
        raise FileNotFoundError(f"Splits dictionary not found in {splits_file}")

    data_splits = {}
    with open(splits_file, "r") as f:
        splits_data = json.load(f)

    for split_name in predefined_split_info["splits"]:
        if "val" in split_name and not use_validation:
            data_splits["val"] = None
            continue
        split_ids = splits_data[split_name]
        data_splits[split_name] = data.filter(lambda x: x["id"] in split_ids)

    return data_splits.values()  # data_splits["train"], data_splits["test"], data_splits["val"]


def _unpickle(file: Path) -> Any:
    try:
        with open(file, "rb") as f:
            return pkl.load(f)
    except pkl.UnpicklingError:
        with gzip.open(file, "rb") as f:
            return pkl.load(f)


def preprocess_data(
    dataset_name: str, test_size=0.15, use_validation=True
) -> Tuple[Dataset, Dataset]:

    dataset_dir = _DATA_DIR / dataset_name
    assert dataset_dir.exists(), f"Dataset {dataset_name} not found in {dataset_dir}"
    sequences = _unpickle(dataset_dir / "sequences.pkl")
    labels = _unpickle(dataset_dir / "labels.pkl")

    data = {"id": [], "sequence": [], "labels": []}
    for key in sequences.keys():
        data["id"].append(key)
        data["sequence"].append("".join(sequences[key]))
        data["labels"].append(labels[key])
        if len(sequences[key]) != len(labels[key]):
            logging.warning(f"Sequence and label lengths do not match for {key}")
    data = Dataset.from_dict(data)  # pd.DataFrame(data)

    # split data into train, val with fixed seed
    train_test_split = data.train_test_split(test_size=test_size, seed=_SEED)
    train_data = train_test_split["train"]
    test_data = train_test_split["test"]
    val_data = None

    if use_validation:
        train_val_split = train_data.train_test_split(test_size=0.15, seed=_SEED)
        train_data = train_val_split["train"]
        val_data = train_val_split["test"]

    return train_data, test_data, val_data


def post_process_data(data: Dataset, max_length: int = None, pos_range=None) -> Dataset:
    # cut sequences to max_length and labels accordingly
    if max_length is not None:
        data = data.map(
            lambda x: {"sequence": x["sequence"][:max_length], "labels": x["labels"][:max_length]},
            batched=False,
        )
    if pos_range is not None:
        # remove samples with too few positive labels or too many positive labels
        pos_frac = np.array([sum(labels) / len(labels) for labels in data["labels"]])
        filter_mask = (pos_frac >= pos_range[0]) & (pos_frac <= pos_range[1])
        data = data.select(np.where(filter_mask)[0])
    return data


def prune(data: Dataset, prune_frac: float = 0.2, random_seed: int = _SEED) -> Dataset:
    # prune a fraction of the data
    rng = np.random.default_rng(random_seed)
    indices = rng.choice(len(data), int(len(data) * (1 - prune_frac)), replace=False)
    return data.select(indices)


def downsample_negatives(
    X: np.ndarray, y: np.ndarray, downsample_ratio: float, random_seed: int = _SEED
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Downsample the vectors in X that have label 0, according to the specified downsample_ratio.

    Args:
    - X (np.ndarray): The input array of vectors (shape: [n_samples, n_features]).
    - y (np.ndarray): The corresponding label array (shape: [n_samples, ]).
    - downsample_ratio (float): The fraction of label 0 vectors to retain (0.0 < downsample_ratio <= 1.0).
    - random_seed (int, optional): A random seed to ensure reproducibility. Defaults to None.

    Returns:
    - X_downsampled (np.ndarray): The downsampled array of vectors.
    - y_downsampled (np.ndarray): The corresponding downsampled labels.
    """
    if downsample_ratio <= 0.0 or downsample_ratio > 1.0:
        raise ValueError("downsample_ratio must be between 0.0 and 1.0")

    # Set random seed for reproducibility
    rng = np.random.default_rng(random_seed)

    # Get the indices of vectors with label 0 and label 1
    indices_negatives = np.where(y == 0)[0]
    indices_positives = np.where(y == 1)[0]

    num_to_keep = int(len(indices_negatives) * downsample_ratio)
    indices_negatives_to_keep = rng.choice(indices_negatives, num_to_keep, replace=False)
    indices_to_keep = np.concatenate([indices_negatives_to_keep, indices_positives])
    np.random.shuffle(indices_to_keep)

    # Return the downsampled X and y
    return X[indices_to_keep], y[indices_to_keep]


def get_class_weights(dataset: Dataset, label_column: str = "labels"):
    labels = np.concatenate(dataset[label_column])
    class_weights = compute_class_weight(
        class_weight="balanced", classes=np.unique(labels), y=labels
    )
    return class_weights
