"""HugingFace models utils"""

from pathlib import Path
import logging
import yaml

import h5py
import torch
from transformers import (
    AutoModelForTokenClassification,
    AutoTokenizer,
    DataCollatorForTokenClassification,
    Trainer,
    TrainingArguments,
    EarlyStoppingCallback,
    EvalPrediction,
    set_seed as set_seed_hf,
)
from transformers.utils import logging as hf_logging
from accelerate.utils import find_executable_batch_size
import peft

from utils import (
    compute_metrics as _compute_metrics,
    set_torch_device,
    process_config_file,
    HF_PRETRAINED_MODELS,
)
# from torch_models import TorchModelHelper
import data_utils


_SEED = 42


def set_seed():
    set_seed_hf(_SEED)


def set_verbosity(verbosity_level: int) -> None:
    if verbosity_level == 0:
        hf_logging.set_verbosity_error()
    elif verbosity_level == 1:
        hf_logging.set_verbosity_warning()
    elif verbosity_level == 2:
        hf_logging.set_verbosity_info()
    else:
        raise ValueError("Verbosity level must be between 0 and 2")


def compute_metrics(eval_pred: EvalPrediction):
    preds, labels = eval_pred
    mask = labels != -100  # ignore padding tokens
    labels = labels[mask]
    preds = preds[mask]
    return _compute_metrics(labels, preds)


class HFModelHelper:
    """Common class for loading `transformers.PreTrainedModel` models"""

    def __init__(self, model_name: str, model_path: str = None, device: str = "cpu"):
        model_id = HF_PRETRAINED_MODELS.get(model_name.upper(), model_name)
        if model_path is None:
            model_path = model_id
        self.model_path = model_path
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_id, use_fast=("rostlab" not in model_id.lower())
        )
        # load base pre-trained model
        try:
            model = AutoModelForTokenClassification.from_pretrained(
                model_path, num_labels=2, classifier_dropout=0.2
            )
            if model.dropout.p != 0.2:
                logging.warning(
                    "Classifier droput was not initialized to 0.2 Setting it manually."
                )
            model.dropout.p = 0.2
            is_peft = False
        except:
            raise ValueError(f"Model {model_path} is not a valid Hugging Face model")
        # load peft adapter if present
        try:
            model = peft.PeftModel.from_pretrained(model, model_id=model_path)
            is_peft = True
        except ValueError:
            if Path(model_path).is_dir():
                logging.warning(f"PEFT adapter not found at {model_path}. Loading base model.")

        self.model = model
        self.device = device
        self.is_peft = is_peft

        # Freeze all layers except the classifier for non-PEFT models
        if not self.is_peft:
            for name, param in self.model.named_parameters():
                if "classifier" not in name:
                    param.requires_grad = False

    @property
    def device(self):
        return next(self.model.parameters()).device

    @device.setter
    def device(self, device: str):
        self.model.to(device)

    def get_model(self):
        return self.model

    def get_tokenizer(self):
        return self.tokenizer

    def get_collator(self):
        return DataCollatorForTokenClassification(tokenizer=self.tokenizer)

    def tokenize(self, sequences, **kwargs):
        if "saprot" in self.tokenizer.name_or_path.lower():
            # add mask token for structure vocabulary
            if isinstance(sequences, str):
                sequences = "#".join(sequences)
            elif isinstance(sequences, list):
                sequences = ["#".join(seq) for seq in sequences]
        elif "rostlab" in self.tokenizer.name_or_path.lower():
            if isinstance(sequences, str):
                sequences = " ".join(sequences)
            elif isinstance(sequences, list):
                sequences = [" ".join(seq) for seq in sequences]
        return self.tokenizer(sequences, **kwargs)

    def convert_to_peft(self):
        logging.info("Adding LoRA layers to model")
        if self.is_peft:
            logging.warning("Model is already a PEFT model. Skipping conversion.")
            return
        lora_config = peft.LoraConfig(
            task_type=peft.TaskType.TOKEN_CLS,
            r=16,
            base_model_name_or_path=self.model_path,
            target_modules=["q_proj", "v_proj", "query", "value", "q", "v"],
            lora_dropout=0.2,
            lora_alpha=32,
        )
        lora_model = peft.get_peft_model(self.model, lora_config)
        lora_model.print_trainable_parameters()
        self.model = lora_model
        self.is_peft = True


class BalancedTrainer(Trainer):
    """Custom Trainer class for handling class imbalance in token classification tasks"""

    def __init__(self, class_weights, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.class_weights = class_weights

    def compute_loss(self, model, inputs, return_outputs=False):
        """
        How the loss is computed by Trainer. By default, all models return the loss in
        the first element.
        """
        if self.label_smoother is not None and "labels" in inputs:
            labels = inputs.pop("labels")
        else:
            labels = None
        outputs = model(**inputs)

        if labels is not None:
            loss = self.label_smoother(outputs, labels)
        else:

            # Changes start here
            # loss = outputs["loss"] if isinstance(outputs, dict) else outputs[0]
            logits = outputs["logits"]
            criterion = torch.nn.CrossEntropyLoss(weight=self.class_weights)
            loss = criterion(logits.view(-1, 2), inputs["labels"].view(-1))
            # Changes end here

        return (loss, outputs) if return_outputs else loss


def process_dataset(dataset, config, model_helper, split: str = "train"):
    max_length = config["dataset"].get("max_length")
    pos_range = config["dataset"].get("pos_range") if split == "train" else None

    logging.info(f"Processing {split} data")
    if max_length is not None:
        max_length = max_length - 2  # account for special tokens
    dataset = data_utils.post_process_data(dataset, max_length=max_length, pos_range=pos_range)

    # prune train dataset if specified
    prune_frac = config["dataset"].get("prune_frac")
    if prune_frac is not None and split == "train":
        logging.info(f"Pruning {prune_frac} of the data")
        dataset = data_utils.prune(dataset, prune_frac=prune_frac)

    def encode(sample):
        return {
            **model_helper.tokenize(
                sample["sequence"], max_length=max_length, truncation=split != "test"
            )
        }

    # tokenize sequences
    logging.info(f"Tokenizing {split} data")
    dataset = dataset.map(encode, batched=False)

    dataset.set_format(type="torch", columns=["input_ids", "attention_mask", "labels"])
    logging.warning(f"Dataset processed: {split} samples: {len(dataset)}")
    return dataset


def get_class_weights(dataset, device="cpu"):
    class_weights = data_utils.get_class_weights(dataset)
    return torch.tensor(class_weights, device=device, dtype=torch.float)


@find_executable_batch_size(starting_batch_size=64)
def batch_finder(batch_size, model_helper, max_length):
    """Simulate training step to find the maximum batch size that fits into GPU memory"""
    assert (
        torch.device(model_helper.device).type == "cuda"
    ), "Batch finder only works with CUDA devices"
    max_token_id = model_helper.get_tokenizer().vocab_size
    if max_length is None:
        max_length = 1024
    dummy_input_ids = torch.randint(
        0, max_token_id, (batch_size, max_length), device=model_helper.device
    )
    labels = torch.randint(0, 2, (batch_size, max_length), device=model_helper.device)
    loss = model_helper.model(input_ids=dummy_input_ids, labels=labels).loss
    loss.backward()

    return batch_size // 2


def load_data(config, model_helper):
    # load and process data
    predefined_split_info = config["dataset"].get("predefined_split_info", False)
    train_data, test_data, val_data = data_utils.load_data(
        dataset_name=config["dataset"]["name"],
        predefined_split_info=predefined_split_info,
        test_size=0.15,
    )
    train_data = process_dataset(train_data, config, model_helper, split="train")
    val_data = process_dataset(val_data, config, model_helper, split="validation")
    test_data = process_dataset(test_data, config, model_helper, split="test")
    return train_data, val_data, test_data


def get_model_helper(model_name, config):
    device = set_torch_device()
    if model_name.upper() in HF_PRETRAINED_MODELS:
        logging.info(f"Loading Hugging Face model: {model_name}")
        model_helper = HFModelHelper(model_name, device=device)
        if config["model"]["is_peft"]:
            model_helper.convert_to_peft()
    # else:
    #     logging.info(f"Loading custom torch model: {model_name}")
    #     model_config = config["model"]
    #     model_config.pop("model_name", None)
    #     try:
    #         pretrained_model = model_config.pop("pretrained_model")
    #     except KeyError:
    #         raise ValueError(
    #             "Pretrained model for downstrem torch model not found in configuration file"
    #         )
    #     model_helper = TorchModelHelper(model_name, pretrained_model, device, **model_config)
    logging.debug(f"Model: {model_helper.model}")
    return model_helper


def get_training_arguments(train_config, model_helper, max_length, logging_dir):
    train_config["logging_dir"] = logging_dir
    train_config["output_dir"] = logging_dir

    # # Warm-up run to initialize CUDA context
    # logging.info("Performing warm-up run to initialize CUDA context")
    # try:
    #     batch_finder(batch_size=1, model_helper=model_helper, max_length=max_length)
    # except RuntimeError as e:
    #     logging.warning(f"Warm-up run failed: {e}")

    # set batch size
    logging.info("Finding maximum instantaneous batch size")
    effective_batch_size = train_config.get("per_device_train_batch_size", 8)
    max_train_batch_size = batch_finder(model_helper=model_helper, max_length=max_length)
    train_batch_size = min(effective_batch_size, max_train_batch_size)
    eval_batch_size = train_batch_size  # * 4
    gradient_accumulation_steps = max(1, effective_batch_size // train_batch_size)
    logging.info(
        f"Batch size: Eval: {eval_batch_size}, train: {train_batch_size},"
        f" gradient accumulation: {gradient_accumulation_steps}"
    )
    train_config["per_device_train_batch_size"] = train_batch_size
    train_config["per_device_eval_batch_size"] = eval_batch_size
    train_config["gradient_accumulation_steps"] = gradient_accumulation_steps
    train_config["learning_rate"] = (
        train_config.get("learning_rate", 5e-5) * gradient_accumulation_steps
    )

    train_args = TrainingArguments(**train_config)
    return train_args


def train(
    model_name: str,
    config_file: str,
    logging_dir: str = "./log",
    verbosity: int = 2,
):
    """Train function for Huggingface models"""

    set_verbosity(verbosity)
    set_seed()

    # load config
    config = process_config_file(config_file, logging_dir)
    max_length = config["dataset"].get("max_length", 1024)

    # load model and data
    model_helper = get_model_helper(model_name, config)
    train_data, val_data, test_data = load_data(config, model_helper)

    # initialize Trainer
    class_weights = get_class_weights(train_data, device=model_helper.device)
    train_config = config["trainer"]
    train_args = get_training_arguments(train_config, model_helper, max_length, logging_dir)
    logging.info(f"Class weights: {class_weights}")
    logging.debug(f"Training arguments: {train_args}")

    trainer = BalancedTrainer(
        class_weights=class_weights,
        model=model_helper.model,
        args=train_args,
        data_collator=model_helper.get_collator(),
        train_dataset=train_data,
        eval_dataset=val_data,
        callbacks=[EarlyStoppingCallback(early_stopping_patience=2)],
        compute_metrics=compute_metrics,
    )

    train_output = trainer.train()
    train_metrics = trainer.metrics_format(train_output.metrics)
    with open(f"{logging_dir}/train_metrics.yml", "a+") as file:
        yaml.dump(train_metrics, file, default_flow_style=False)

    # evaluate on test data
    logging.info("Evaluating on test data")
    test_metrics = trainer.evaluate(test_data)
    with open(f"{logging_dir}/test_metrics.yml", "a+") as file:
        yaml.dump(trainer.metrics_format(test_metrics), file, default_flow_style=False)
    trainer.log_metrics("test", test_metrics)

    best_model_checkpoint = trainer.state.best_model_checkpoint
    return best_model_checkpoint


def predict(checkpoint_dir, output_file: str, verbosity: int = 2):
    """Predict function for Huggingface models"""

    set_verbosity(verbosity)
    device = set_torch_device()

    test_data = 1
    # load model
    model_helper = HFModelHelper(checkpoint_dir, device)
    args = TrainingArguments(per_device_eval_batch_size=1)

    # For predictions, no need to use balanced trainer
    trainer = Trainer(
        model=model_helper.model,
        args=args,
        data_collator=model_helper.get_collator(),
        compute_metrics=compute_metrics,
    )

    # compute predictions
    preds = trainer.predict(test_data).predictions

    # save predictions
    Path(output_file).parent.mkdir(parents=True, exist_ok=True)
    with h5py.File(output_file, "w") as f:
        for i, sample in enumerate(test_data):
            id = sample["id"]
            f.create_dataset(id, data=preds[i])
