"""Hyperparameter search for HuggingFace models using Optuna backend."""

from typing import Tuple, Callable, Any
from pathlib import Path
import os
import sys
import yaml
import logging
import importlib.util

from transformers import AutoTokenizer, AutoModelForTokenClassification, EarlyStoppingCallback
import peft

sys.path.append(Path(__file__).parent)
import utils
import hf_utils


class HPOModelHelper(hf_utils.HFModelHelper):
    """
    Provide access to hyperparameters initialization
    and direct conversion to peft model.
    """

    def __init__(
        self,
        model_name: str,
        model_path: str = None,
        device: str = "cpu",
        dropout: float = 0.2,
        lora_r: int = 16,
        lora_target_modules: list = ["q_proj", "v_proj", "query", "value", "q", "v"],
    ):
        model_id = utils.HF_PRETRAINED_MODELS.get(model_name.upper(), model_name)
        if model_path is None:
            model_path = model_id
        self.model_path = model_path
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_id, use_fast=("rostlab" not in model_id.lower())
        )
        # load base pre-trained model
        model = AutoModelForTokenClassification.from_pretrained(
            model_path, num_labels=2, classifier_dropout=dropout
        )
        if model.dropout.p != dropout:
            logging.warning(
                f"Classifier droput was not initialized to {dropout}. Setting it manually."
            )
        model.dropout.p = dropout

        # load peft adapter if present
        is_peft = False
        try:
            model = peft.PeftModel.from_pretrained(model, model_id=model_path)
            is_peft = True
        except ValueError:
            logging.warning(
                f"PEFT adapter not found at {model_path}. Loading base model and initializing lora"
                " layers."
            )

        if not is_peft:
            lora_config = peft.LoraConfig(
                task_type=peft.TaskType.TOKEN_CLS,
                r=lora_r,
                base_model_name_or_path=self.model_path,
                target_modules=lora_target_modules,
                lora_dropout=dropout,
                lora_alpha=2 * lora_r,
            )
            lora_model = peft.get_peft_model(model, lora_config)
            lora_model.print_trainable_parameters()
            self.model = lora_model
        else:
            self.model = model

        self.device = device

    def convert_to_peft(self):
        raise NotImplementedError(
            "Conversion to PEFT model is already done during initialization."
        )


def set_wandb_env(logging_dir: str):
    if importlib.util.find_spec("wandb") is not None:
        project = logging_dir.split("hflogs/")[-1].replace("/", "-")
        os.environ["WANDB_PROJECT"] = project
        os.environ["WANDB_WATCH"] = "parameters"
        logging.info("wandb is available and environment variables are set.")
    else:
        logging.info("wandb is not available.")


def _prepare_hpo(model_name: str, hpo_config: dict) -> Tuple[Callable[..., Any]]:
    hyperparams = hpo_config["hyperparameters"]
    objective = hpo_config.get("objective", "roc_auc")
    logging.info(f"Objective metric: {objective}")

    def optuna_hp_space(trial):
        learning_rate = trial.suggest_float(
            "learning_rate", **hyperparams["learning_rate"], log=True
        )
        weight_decay = trial.suggest_categorical("weight_decay", hyperparams["weight_decay"])
        warmup_ratio = trial.suggest_categorical("warmup_ratio", hyperparams["warmup_ratio"])
        params = {
            "learning_rate": learning_rate,
            "weight_decay": weight_decay,
            "warmup_ratio": warmup_ratio,
        }
        logging.info(f"Trial hyperparameters: {params}")
        return params

    def compute_objective(metrics):
        return metrics["eval_" + objective]

    def model_init(trial):
        """Initialize the model."""
        # In the first initialization, trial is None and default values are used to avoid errors
        # logging.info("Initializing model during trial")
        dropout = (
            trial.suggest_categorical("dropout", hyperparams["dropout"])
            if "dropout" in hyperparams and trial is not None
            else 0.2
        )
        lora_r = (
            trial.suggest_categorical("lora_r", hyperparams["lora_r"])
            if "lora_r" in hyperparams and trial is not None
            else 16
        )
        lora_target_modules = (
            trial.suggest_categorical("lora_target_modules", hyperparams["lora_target_modules"])
            if "lora_target_modules" in hyperparams and trial is not None
            else ["q_proj", "v_proj", "query", "value", "q", "v"]
        )
        # since optuna `suggest_categorical requires choices to be hashable, in the config file
        # the choices are defined as strings separated by "__"
        if isinstance(lora_target_modules, str):
            lora_target_modules = lora_target_modules.split("__")
        params = {"dropout": dropout, "lora_r": lora_r, "lora_target_modules": lora_target_modules}
        if trial is not None:
            logging.info(f"Model hyperparameters: {params}")
        model_helper = HPOModelHelper(
            model_name, **params
        )  # may be necessary to add trial arguments
        return model_helper.get_model()

    return model_init, optuna_hp_space, compute_objective


def run_hpo(
    model_name: str,
    config_file: str,
    logging_dir: str = "./log",
    verbosity: int = 2,
):
    hf_utils.set_seed()
    hf_utils.set_verbosity(verbosity)
    device = utils.set_torch_device()

    # process config file and copy to logging directory
    Path(logging_dir).mkdir(parents=True, exist_ok=True)
    config = utils.process_config_file(config_file, logging_dir)
    max_length = config["dataset"].get("max_length", 1024)

    # initlialize base model helper for loading data and computing class weights
    model_helper = hf_utils.HFModelHelper(model_name, device=device)
    model_helper.convert_to_peft()
    train_data, val_data, _ = hf_utils.load_data(config, model_helper)
    print(train_data[0])
    class_weights = hf_utils.get_class_weights(train_data, device=device)
    data_collator = model_helper.get_collator()

    # get default training arguments
    set_wandb_env(logging_dir)
    train_config = config["trainer"]
    patience = train_config.pop("early_stopping_patience", None)
    train_args = hf_utils.get_training_arguments(
        train_config, model_helper, max_length, logging_dir
    )
    del model_helper  # remove model helper to free up memory

    # get hpo util functions
    hpo_config = config["hpo"]
    model_init, optuna_hp_space, compute_objective = _prepare_hpo(model_name, hpo_config)

    # initialize trainer
    trainer = hf_utils.BalancedTrainer(
        class_weights=class_weights,
        model=None,
        model_init=model_init,
        args=train_args,
        data_collator=data_collator,
        train_dataset=train_data,
        eval_dataset=val_data,
        callbacks=[EarlyStoppingCallback(early_stopping_patience=patience)] if patience else None,
        compute_metrics=hf_utils.compute_metrics,
    )

    # run hyperparameter search
    logging.info("Starting hyperparameter search...")
    best_trial = trainer.hyperparameter_search(
        direction="maximize",
        hp_space=optuna_hp_space,
        compute_objective=compute_objective,
        n_trials=hpo_config.get("n_trials", 20),
        backend="optuna",
    )

    # Save the best trial hyperparameters to a YAML file
    result = dict()
    result["run_id"] = best_trial.run_id
    result["roc_auc"] = best_trial.objective
    result["hyperparameters"] = best_trial.hyperparameters
    output_file = Path(logging_dir) / "best_trial.yaml"
    with open(output_file, "w") as file:
        yaml.dump(result, file, default_flow_style=False)
    logging.info(f"Best trial hyperparameters saved to {output_file}")


if __name__ == "__main__":
    args = utils.parse_cli(__doc__)
    utils.setup_logging(args.verbosity)
    logging.info(f"Starting hyperparameter search for model: {args.model_name}")
    run_hpo(args.model_name, args.config_file, args.logging_dir, args.verbosity)
