"""Utils functions for setup, logging and training."""

from pathlib import Path
import logging
import yaml
import shutil

from sklearn import metrics
from scipy.special import softmax


def setup_logging(verbosity: int) -> None:
    if verbosity == 2:
        level = logging.DEBUG
    elif verbosity == 1:
        level = logging.INFO
    elif verbosity == 0:
        level = logging.WARNING
    logging.basicConfig(
        level=level,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        handlers=[logging.StreamHandler()]
    )
    

def parse_cli(description):
    import argparse
    
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-m",
        "--model-name",
        type=str,
        required=True,
        help=(
            "The name of the model to train. Must be a valid Hugging Face model, a local"
            " checkpoint path, or 'sklearn'."
        ),
    )
    parser.add_argument(
        "-c",
        "--config-file",
        type=str,
        required=True,
        help="The path to the configuration file.",
    )
    parser.add_argument(
        "-o",
        "--logging-dir",
        type=str,
        default="./log",
        help="The directory where logs and model checkpoints will be saved.",
    )
    parser.add_argument(
        "--pred-file",
        type=str,
        default=None,
        help="The path to the file where the predictions will be saved.",
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        type=int,
        default=2,
        help="The verbosity level of the logger.",
    )
    args = parser.parse_args()

    # Check for invalid arguments and set defaults
    if args.verbosity > 2:
        logging.warning("Verbosity level must be between 0 and 2. Setting to 2.")
        args.verbosity = 2
    elif args.verbosity < 0:
        logging.warning("Verbosity level must be between 0 and 2. Setting to 0.")
        args.verbosity = 0

    return args


def process_config_file(config_file, logging_dir, sanity_check_fn=None):
    # load config file
    with open(config_file, "r") as file:
        config = yaml.safe_load(file)
    # run sanity check function
    if sanity_check_fn:
        sanity_check_fn(config)
    # remove *.yml files from logging directory
    for file in Path(logging_dir).glob("*.yml"):
        file.unlink()
    shutil.copy2(config_file, logging_dir)
    return config


# Pretrained models hub IDs
HF_PRETRAINED_MODELS = {
    "ESM2_35M": "facebook/esm2_t12_35M_UR50D",
    "ESM2_650M": "facebook/esm2_t33_650M_UR50D",
    "PROTT5": "Rostlab/prot_t5_xl_uniref50",
    "ANKH_BASE": "ElnaggarLab/ankh-base",
    "ANKH_LARGE": "ElnaggarLab/ankh-large",
    "SAPROT_35M": "westlake-repl/SaProt_35M_AF2",
    "SAPROT_650M": "westlake-repl/SaProt_650M_AF2",
}


_DEAFAULT_EVAL_METRICS = {
    "f1": metrics.f1_score,
    "mc": metrics.matthews_corrcoef,
    "balanced_accuracy": metrics.balanced_accuracy_score,
    "average_precision": metrics.average_precision_score,
    "precision": metrics.precision_score,
    "recall": metrics.recall_score,
    "roc_auc": metrics.roc_auc_score,
    "accuracy": metrics.accuracy_score,
}


def compute_metrics(labels, preds):
    eval_results = {}
    for metric_name, metric_func in _DEAFAULT_EVAL_METRICS.items():
        try:
            eval_results[metric_name] = metric_func(labels, softmax(preds, axis=-1)[:, 1])
        except ValueError:
            try:
                eval_results[metric_name] = metric_func(labels, preds.argmax(axis=-1))
            except Exception as e:
                raise e
    return eval_results


def set_torch_device():
    import torch

    if torch.cuda.is_available():
        device = "cuda"
    elif torch.backends.mps.is_available():
        device = "mps"
    else:
        device = "cpu"
    logging.info(f"Using default device: {device}")
    return device
